#include <cv.h>
#include <cxcore.h>
#include <highgui.h>
#include <vector>
#include <cstdio>
#include <cstring>
#include <iostream>

#include <lbp_test.h> // the classifier code

using namespace std;


int eval(unsigned char * data, int ls, float & response)
{
    response = 0;
    
    EVAL_test_CLASSIFIER(data,ls,response);

    return (response > 0);
}


void scanImage(IplImage * image, vector<CvPoint> & detections)
{
    char * row = image->imageData;
    int y = 0;
    float response;
    detections.clear();
    detections.reserve(512);
    while (row < image->imageData + (image->height-24) * image->widthStep)
    {
        char * px = row;
        while (px < row + (image->width-24))
        {
            if (eval((unsigned char*)px, image->widthStep, response))
            {
                detections.push_back(cvPoint(px-row, y));
            }
            px+=1;
        }
        ++y;
        row += image->widthStep;
    }
}


int main()
{
    CvCapture * cap = cvCaptureFromCAM(0);

    if (!cap)
    {
        cerr << "You have no camera lol." << endl;
        return 0;
    }

    IplImage * f_gray = 0;
    IplImage * f_small = 0;
    IplImage * f_mask = 0;
    bool init = false;

    cvNamedWindow("IMAGE",1);
    cvNamedWindow("MASK",1);

    vector<CvPoint> dets;
    
    while (1)
    {
        IplImage * img = cvQueryFrame(cap);

        if (!init)
        {
            f_small = cvCreateImage(cvSize(320,240), IPL_DEPTH_8U, 3);
            f_gray = cvCreateImage(cvGetSize(f_small), IPL_DEPTH_8U, 1);
            f_mask = cvCreateImage(cvGetSize(f_small), IPL_DEPTH_8U, 1);
            init = true;
        }
        cvSmooth(img, img, CV_GAUSSIAN, 5);
        cvResize(img, f_small);

        cvCvtColor(f_small, f_gray, CV_RGB2GRAY);
        
        scanImage(f_gray, dets); // actual detection

        cvZero(f_mask);
        for (unsigned i = 0; i < dets.size(); ++i)
        {
            CvPoint pt = dets[i];
            *(f_mask->imageData + (pt.y+12)*f_mask->widthStep + (pt.x+12)) = 255;
        }

        cvShowImage("IMAGE", f_gray);
        cvShowImage("MASK", f_mask);

        char c = cvWaitKey(50);
        if (c == 'q') break;
    }

    cvReleaseImage(&f_gray);
    cvReleaseImage(&f_mask);
    cvReleaseCapture(&cap);

    return 0;
}

