//
//  features.h
//  Macros for static evaluation of LRP and LBP features
//  $Id$
//

#ifndef FEATURE_MACROS
#define FEATURE_MACROS


// RELEASE==1 produces C++ compilant code
// RELEASE==0 produces simplified code where [x,y] means access to an image pixel
#define RELEASE 1

#if RELEASE==1
#define px11(data,x,y,ls) *((data)+(ls)*(y)+(x))
#else
#define px11(data,x,y,ls)  [x,y]
#endif

#define px21(data,x,y,ls)  (px11(data,x,y,ls)+px11(data,(x)+1,y,ls))
#define px12(data,x,y,ls)  (px11(data,x,y,ls)+px11(data,x,(y)+1,ls))
#define px22(data,x,y,ls)  (px21(data,x,y,ls)+px21(data,x,(y)+1,ls))

#define CMP(x,y) ((x)>(y))

#define BIT(x,b) ((x)<<(b)) // valid only for x==0 or x==1 !!!


#define COUNT(data,ls,base,xs,ys,PX)\
  (CMP(PX(data,0,0,ls),base)+CMP(PX(data,xs,0,ls),base)+CMP(PX(data,2*xs,0,ls),base)+\
   CMP(PX(data,0,ys,ls),base)+CMP(PX(data,xs,ys,ls),base)+CMP(PX(data,2*xs,ys,ls),base)+\
   CMP(PX(data,0,2*ys,ls),base)+CMP(PX(data,xs,2*ys,ls),base)+CMP(PX(data,2*xs,2*ys,ls),base))


#define LRP(data,ls,x,y,Ax,Ay,Bx,By,xs,ys,PX,alpha,response)\
    {\
        unsigned char * d=data+y*ls+x;\
        int A=PX(d,Ax*xs,Ay*ys,ls);\
        int B=PX(d,Bx*xs,By*ys,ls);\
        int lrp=10*COUNT(d,ls,A,xs,ys,PX)+COUNT(d,ls,B,xs,ys,PX);\
        response+=*((float*)(alpha)+lrp);\
    }

#if RELEASE==1
#define LBP(data,ls,x,y,xs,ys,PX,alpha,response)\
{\
    unsigned char * d=data+y*ls+x;\
    int base=PX(d,1*xs,1*ys,ls);\
    int lbp=\
    BIT(CMP(PX(d,0*(xs),0*(ys),ls),base),0) | \
    BIT(CMP(PX(d,1*(xs),0*(ys),ls),base),1) | \
    BIT(CMP(PX(d,2*(xs),0*(ys),ls),base),2) | \
    BIT(CMP(PX(d,0*(xs),1*(ys),ls),base),7) | \
    BIT(CMP(PX(d,2*(xs),1*(ys),ls),base),3) | \
    BIT(CMP(PX(d,0*(xs),2*(ys),ls),base),6) | \
    BIT(CMP(PX(d,1*(xs),2*(ys),ls),base),5) | \
    BIT(CMP(PX(d,2*(xs),2*(ys),ls),base),4);\
    response+=*((alpha)+lbp);\
}
#else
#define LBP(data,ls,x,y,xs,ys,PX,alpha,response)\
{\
    base=PX(d,xs,ys,ls)\
    p0=CMP(PX(d,0,0,ls),base)\
    p1=CMP(PX(d,xs,0,ls),base)\
    p2=CMP(PX(d,xs+xs,0,ls),base)\
    p3=CMP(PX(d,0,ys,ls),base)\
    p4=CMP(PX(d,xs+xs,ys,ls),base)\
    p5=CMP(PX(d,0,ys+ys,ls),base)\
    p6=CMP(PX(d,xs,ys+ys,ls),base)\
    p7=CMP(PX(d,xs+xs,ys+ys,ls),base)\
    lbp=BIT(p0,0) | BIT(p1,1) |BIT(p2,2) | BIT(p3,7) |BIT(p4,3) | BIT(p5,6) |BIT(p6,5) | BIT(p7,4)\
}
#endif

#endif
