/*
 * classifier.cpp
 * $Id: classifier.cpp 178 2009-10-08 12:21:51Z ijuranek $
 *
 * Notes:
 */

#include <classifier.h>
#include <simplexml.h>
#include <sstream>
#include <iostream>
#include <map>
#include <stack>
#include <vector>
#include <list>
#include <cassert>

using namespace std;

int initClassifier(TClassifier * classifier)
{
    float * stgAlpha = classifier->alpha;

    //printf("classifier=%08X; tp=%d\n", classifier, classifier->tp);
    //printf("stages=%d; alphas=%d; sz=%dx%d\n", classifier->stageCount, classifier->alphaCount, classifier->width, classifier->height);
    
    for (unsigned s = 0; s < classifier->stageCount; ++s, stgAlpha += classifier->alphaCount)
    {
        if (classifier->tp == LRD)
        {
            TStage & stage = ((TStage*)(classifier->stage))[s];
            /*
            if (stage.w > 2 || stage.h > 2) // no blocks larger than 2x2 allowed
            {
                return 0;
            }
            */

            // additional parameter checks?

            // get feature type
            // 000000wh
            stage.szType = ((stage.h-1) << 1) | (stage.w-1);
            // get position type
            // 0000yyxx
            stage.posType = ((stage.y & 0x03) << 2) | (stage.x & 0x03);
            // link alpha table
            stage.alpha = stgAlpha + 8; // offset the pointer to avoid sum later in feature evaluation

            /*printf("%d: x=%d; y=%d; w=%d; h=%d; a=%d; b=%d; theta=%f; alpha=%d\n",\
                    s, stage.x, stage.y, stage.w, stage.h, stage.a, stage.b, stage.theta_b, stage.alpha - classifier->alpha);*/

            continue;
        }
        
        // Essentially the same as the LRD classifier but with few changes
        if (classifier->tp == LRP || classifier->tp == LBP)
        {
            TStage & stage = ((TStage*)(classifier->stage))[s];
            
            /*
            if (stage.w > 2 || stage.h > 2) // no blocks larger than 2x2 allowed
            {
                return 0;
            }
            */

            stage.szType = ((stage.h-1) << 1) | (stage.w-1);
            stage.posType = ((stage.y & 0x03) << 2) | (stage.x & 0x03);
            stage.alpha = stgAlpha;

            /*printf("%d: x=%d; y=%d; w=%d; h=%d; a=%d; b=%d; theta=%f; alpha=%d\n",\
                    s, stage.x, stage.y, stage.w, stage.h, stage.A, stage.B, stage.theta_b, stage.alpha - classifier->alpha);*/

            continue;
        }

        if (classifier->tp == HAAR)
        {
            THaarStage & stage = ((THaarStage*)(classifier->stage))[s];
            stage.alpha = stgAlpha;
            continue;
        }

    }

    return 1;
}

template<class S>
static inline void disableWB(S *s, const unsigned c)
{
    for(unsigned i = 0; i < c; ++i)
    {
        s[i].theta_b = -1e+010;
    }
}

void disableClassifierWaldBoost(TClassifier * c)
{
    assert(c != NULL);

    switch(c->tp)
    {
        case LRD : disableWB((TStage*)(c->stage), c->stageCount); break;
        case LRP : disableWB((TStage*)(c->stage), c->stageCount); break;
        case LRP3: disableWB((TStage*)(c->stage), c->stageCount); break;
        case LBP : disableWB((TStage*)(c->stage), c->stageCount); break;
        case HAAR : disableWB((THaarStage*)(c->stage), c->stageCount); break;
        default : assert(!"Classifier type not supported!");
    };
}

void releaseClassifier(TClassifier ** classifier)
{
    if (classifier && *classifier)
    {
        TClassifier & c = **classifier;
        if (c.alpha) delete [] c.alpha;
        
        switch (c.tp)
        {
        case LRD:
        case LRP:
        case LRP3:
        case LBP:
            if (c.stage) delete [] (TStage*)c.stage;
            break;
        case HAAR:
            if (c.stage) delete [] (THaarStage*)c.stage;
            break;
        default:
            break;
        };
        
        delete *classifier;
        *classifier = 0;
    }
}


static void loadLRDFeature(xmlNodePtr fNode, TStage * stage)
{
    if (!fNode) return;
    xmlNodePtr lrNode = 0;
    if (!lrNode)
    {
        lrNode = getNode("LRDFeature", fNode);
    }
    if (!lrNode)
    {
        lrNode = getNode("LRP", fNode);
    }
    if (!lrNode) return;

    getAttr(stage->x, "positionX", lrNode);
    getAttr(stage->y, "positionY", lrNode);
    getAttr(stage->w, "blockWidth", lrNode);
    getAttr(stage->h, "blockHeight", lrNode);
    int A,B,C;
    getAttr(A, "blockA", lrNode);
    getAttr(B, "blockB", lrNode);
    getAttr(C, "blockC", lrNode);
    stage->A = A;
    stage->B = B;
    stage->C = C;
}

static void loadLBPFeature(xmlNodePtr fNode, TStage * stage)
{
    if (!fNode) return;
    xmlNodePtr lrNode = 0;
    if (!lrNode)
    {
        lrNode = getNode("LBPFeature", fNode);
    }
    if (!lrNode) return;

    getAttr(stage->x, "positionX", lrNode);
    getAttr(stage->y, "positionY", lrNode);
    getAttr(stage->w, "blockWidth", lrNode);
    getAttr(stage->h, "blockHeight", lrNode);
}


static void loadHaarFeature(xmlNodePtr fNode, THaarStage * stage)
{
    xmlNodePtr discFeature = getNode("TCont2DiscFeature", fNode);
    if (!discFeature)
    {
        return;
    }

    getAttr(stage->min, "minValue", discFeature);
    getAttr(stage->max, "maxValue", discFeature);
    getAttr(stage->bins, "numberOfBins", discFeature);


    xmlNodePtr haarPtr;
    haarPtr = getNode("HaarHorizontalDoubleFeature", discFeature);
    if (haarPtr)
    {
        stage->tp = H2V1;
        goto load_params; // bloody fucking goto! I am not proud of myself...
    }
    haarPtr = getNode("HaarVerticalDoubleFeature", discFeature);
    if (haarPtr)
    {
        stage->tp = H1V2;
        goto load_params;
    }
    haarPtr = getNode("HaarHorizontalTernalFeature", discFeature);
    if (haarPtr)
    {
        stage->tp = H3V1;
        goto load_params;
    }
    haarPtr = getNode("HaarVerticalTernalFeature", discFeature);
    if (haarPtr)
    {
        stage->tp = H1V3;
        goto load_params;
    }
    haarPtr = getNode("HaarDiagonalFeature", discFeature);
    if (haarPtr)
    {
        stage->tp = H2V2;
        goto load_params;
    }
    haarPtr = getNode("HaarSurroundFeature", discFeature);
    if (haarPtr)
    {
        stage->tp = SRND;
        goto load_params;
    }

    return;
  
load_params:
    getAttr(stage->x, "positionX", haarPtr);
    getAttr(stage->y, "positionY", haarPtr);
    getAttr(stage->w, "blockWidth", haarPtr);
    getAttr(stage->h, "blockHeight", haarPtr);
    stage->sz = stage->w * stage->h;
}


static void loadHistogramWeakHypothesis(xmlNodePtr hNode, vector<float> & predictTable)
{
    if (!hNode) return;
            
    istringstream predictStr(getAttr("predictionValues", hNode));

    float p;
    while (predictStr >> p)
    {
        predictTable.push_back(p);
    }
}


static void loadDTWeakHypothesis(xmlNodePtr hNode, vector<float> & predictTable)
{
    if (!hNode) return;
            
    istringstream predictStr(getAttr("predictionValues", hNode));
    istringstream binMapStr(getAttr("binMap", hNode));

    vector<float> predict;
    float p;
    while (predictStr >> p)
    {
        predict.push_back(p);
    }

    int bin;
    while (binMapStr >> bin)
    {
        predictTable.push_back(predict[bin]);
    }

}


void loadLRDClassifier(xmlNodePtr stages, TClassifier * classifier);
void loadHaarClassifier(xmlNodePtr stages, TClassifier * classifier);

// loadClassifierXML(const char*)
// findClassifierNode(xmlNodePtr)
// determineClassifierType(xmlNodePtr)
// loadClassifier(xmlNodePtr)

TClassifier * loadClassifier(xmlNodePtr);


xmlNodePtr findClassifierNode(xmlNodePtr rootPtr)
{
    /*
    if (xmlStrcmp(rootPtr->name, BAD_CAST "WaldBoostClassifier") == 0) return rootPtr;
    //xmlNodePtr node = getNode("WaldBoostClassifier", rootPtr);
    //if (!node)
    //{
        xmlNodePtr node = 0;
        for (xmlNodePtr tmp = rootPtr->children; tmp && !node; tmp=tmp->next)
        {
            node = findClassifierNode(tmp);
        }
    //}
    //
    */

    list<xmlNodePtr> classifiers;
    stack<xmlNodePtr> s;
    s.push(rootPtr);
    while (!s.empty())
    {
        xmlNodePtr current = s.top();
        s.pop();
        //const char * name = (const char*) current->name;
        if (xmlStrcmp(current->name, BAD_CAST "WaldBoostClassifier") == 0)
        {
            classifiers.push_back(current);
        }
        for (xmlNodePtr child = current->children; child; child=child->next)
        {
            s.push(child);
        }
    }
    
    if (classifiers.empty())
    {
        return 0;
    }

    if (classifiers.size() > 1)
    {
        cerr << "Warning: The xml contains " << classifiers.size() << " classifiers. Using the first." << endl;
    }

    return classifiers.front();
}


TClassifier * loadClassifierXML(const char * filename)
{
    xmlDocPtr inputXML = xmlParseFile(filename);

    if (!inputXML)
    {
        cerr << "Cannot parse file" << endl;
        return 0;
    }

    //classifierRoot = getNode("WaldBoostClassifier", xmlDocGetRootElement(inputXML));
    xmlNodePtr xmlRoot = xmlDocGetRootElement(inputXML);

    if (!xmlRoot)
    {
        cerr << "Invalid XML" << endl;
        return 0;
    }

    xmlNodePtr classifierNode = findClassifierNode(xmlRoot);

    if (!classifierNode)
    {
        cerr << "Cannot find classifier" << endl;
        return 0;
    }

    return loadClassifier(classifierNode);
}
    
ClassifierType determineClassifierType(xmlNodePtr classifierRoot)
{
    // Try to determine by 'type' attribute

    string tp;
    getAttr(tp, "type", classifierRoot);
    
    if (tp == string("LRD"))
    {
        return LRD;
    }
    if (tp == string("LRP"))
    {
        return LRP;
    }
    if (tp == string("HAAR"))
    {
        return HAAR;
    }
    if (tp == string("LBP"))
    {
        return LBP;
    }

    // No 'type' specified, try to determine by classifier content

    map<string, int> elements;
    // Known feature types
    map<string, ClassifierType> knownTypes;
    knownTypes["LRDFeature"] = LRD;
    knownTypes["LRPFeature"] = LRP;
    knownTypes["LRP3"] = LRP3;
    knownTypes["LBPFeature"] = LBP;
    knownTypes["HaarHorizontalDoubleFeature"] = HAAR;
    knownTypes["HaarVerticalDoubleFeature"] = HAAR;
    knownTypes["HaarHorizontalTernalFeature"] = HAAR;
    knownTypes["HaarVerticalTernalFeature"] = HAAR;
    knownTypes["HaarDiagonalFeature"] = HAAR;
    knownTypes["HaarSurroundFeature"] = HAAR;

    // Enumerate elements used in the classifier
    stack<xmlNodePtr> s;
    s.push(classifierRoot);
    while (!s.empty())
    {
        xmlNodePtr current = s.top();
        s.pop();
        const char * name = (const char*) current->name;
        elements[string(name)]++;
        for (xmlNodePtr child = current->children; child; child=child->next)
        {
            s.push(child);
        }
    }

    ClassifierType type = UNKNOWN;
    bool typeSet = false;
    
    // Go through elements found in the classifier
    for (map<string, int>::iterator i = elements.begin(); i != elements.end(); ++i)
    {
        //cout << i->first << " -> " << i->second << endl;
        if (knownTypes.find(i->first) != knownTypes.end())
        {
            // A supported feature type found
            if (!typeSet)
            {
                // There is no type determined yet
                // the type then corrensponds to the feature type found
                typeSet = true;
                type = knownTypes[i->first];
            }
            else
            {
                if (knownTypes[i->first] != type)
                {
                    // there are mixed feature types while the
                    // framework supprts only monolithic
                    // classifiers
                    // -> fall back to UNKNOWN type
                    type = UNKNOWN;
                }
            }
        }
    }
   
    return type;
}


TClassifier * loadClassifier(xmlNodePtr classifierRoot)
{
    TClassifier * classifier = new TClassifier();

    string tp;
    classifier->threshold = 0.0f;
    getAttr(classifier->width, "sizeX", classifierRoot);
    getAttr(classifier->height, "sizeY", classifierRoot);
    getAttr(classifier->width, "imageSizeX", classifierRoot);
    getAttr(classifier->height, "imageSizeY", classifierRoot);


    classifier->tp = determineClassifierType(classifierRoot);

    //cout << "[classifier: size=" << classifier->width << "x" << classifier->height << "; type=" << tp << "]" << endl;

    switch (classifier->tp)
    {
    case LRD:
        loadLRDClassifier(classifierRoot->children, classifier);
        break;
    case LRP:
        loadLRDClassifier(classifierRoot->children, classifier);
        break;
    case LRP3:
        loadLRDClassifier(classifierRoot->children, classifier);
        break;
    case LBP:
        loadLRDClassifier(classifierRoot->children, classifier);
        break;
    case HAAR:
        loadHaarClassifier(classifierRoot->children, classifier);
        break;
    case UNKNOWN:
    default:
        cerr << "Unknown classifier type!" << endl;
        delete classifier;
        return 0;
        break;
    };
    
    //printf("classifier=%08X, tp=%d\n", classifier, classifier->tp);

    return classifier;
}


// LRD/LRP specific stuff

void loadLRDClassifier(xmlNodePtr stages, TClassifier * classifier)
{
    vector<TStage> tmpStages;
    vector< vector<float> > tmpPredict;
    
    for (xmlNodePtr node = stages; node; node = node->next)
    {
        if (xmlStrcmp(node->name, BAD_CAST "stage") == 0)
        {
            // load stage
            TStage stage = {0,0,1,1,0,1,0.0f};
            //cout << "stage:[";
            
            double theta;
            getAttr(theta, "negT", node);
            if (theta > 5000.0) theta = 5000.0;
            if (theta < -5000.0) theta = -5000.0;
            stage.theta_b = float(theta);
            //cout << "theta=" << stage.theta_b << "; ";

            // load hypothesis part
            xmlNodePtr hypothesisNode = 0;
            if (!hypothesisNode)
            {
                hypothesisNode = getNode("HistogramWeakHypothesis", node);
                if (hypothesisNode)
                {
                    tmpPredict.push_back(vector<float>());
                    loadHistogramWeakHypothesis(hypothesisNode, tmpPredict.back());
                }
                else 
                {
                    hypothesisNode = getNode("DecisionTreeWeakHypothesis", node);
                    if (hypothesisNode)
                    {
                        tmpPredict.push_back(vector<float>());
                        loadDTWeakHypothesis(hypothesisNode, tmpPredict.back());
                    }
                }
            }

            if (!hypothesisNode)
            {
                cerr << "Warning: Incomplete stage encountered" << endl;
            }
            else
            {
                if (classifier->tp == LRD || classifier->tp == LRP)
                    loadLRDFeature(hypothesisNode, &stage);
                if (classifier->tp == LBP)
                    loadLBPFeature(hypothesisNode, &stage);
            }
            
            //cout << stage.x << ";" << stage.y << "]" << endl;
            tmpStages.push_back(stage);
        }
    }

    // Alloc the stages and alphas, and copy loaded data
    
    classifier->stageCount = tmpStages.size();
    classifier->evaluatedStageCount = tmpStages.size();
    classifier->alphaCount = 0;
    if (classifier->tp == LRD) classifier->alphaCount = 17;
    if (classifier->tp == LRP) classifier->alphaCount = 100;
    if (classifier->tp == LRP3) classifier->alphaCount = 1000;
    if (classifier->tp == LBP) classifier->alphaCount = 256;
    assert(classifier->alphaCount > 0 && "Frack! T3h classifier is really weird!");

    // NOTE:
    // alphaCount is restricted to 17 for LRD, 100 for LRP and 256 for LBP
   
    classifier->stage = new TStage[tmpStages.size()];
    classifier->alpha = new float[tmpStages.size() * classifier->alphaCount];

    int errors = 0;

    for (unsigned s = 0; s < tmpStages.size(); ++s)
    {
        ((TStage*)(classifier->stage))[s] = tmpStages[s];
        if (tmpPredict[s].size() != classifier->alphaCount)
        {
            cout << tmpPredict[s].size() << endl;
            ++errors;
        }
        float * dstAlpha = classifier->alpha + classifier->alphaCount * s;

        //TStage & stage = ((TStage*)(classifier->stage))[s];
        //printf("%d: x=%d; y=%d; w=%d; h=%d; a=%d; b=%d; theta=%f",
        //    s, stage.x, stage.y, stage.w, stage.h, stage.A, stage.B, stage.theta_b);

        copy(tmpPredict[s].begin(), tmpPredict[s].end(), dstAlpha);
    }
    if (errors) // W00t, there were errorz?
    {
        cerr << "I am afraid, Dave, I can not load the classifier (" << errors << " errors occured)." << endl;
        releaseClassifier(&classifier);
        return;
    }
    //cout << "Classifier OK" << endl;
}


// HAAR specific stuff

void loadHaarClassifier(xmlNodePtr stages, TClassifier * classifier)
{
    vector<THaarStage> tmpStages;
    vector< vector<float> > tmpPredict;
    
    for (xmlNodePtr node = stages; node; node = node->next)
    {
        if (xmlStrcmp(node->name, BAD_CAST "stage") == 0)
        {
            // load stage
            THaarStage stage = {0,0,1,1,numHaarType,0.0f,0};
            //cout << "stage:[";
            
            double theta;
            getAttr(theta, "negT", node);
            if (theta > 5000.0) theta = 5000.0;
            if (theta < -5000.0) theta = -5000.0;
            stage.theta_b = float(theta);
            //cout << "theta=" << stage.theta_b << "; ";

            // load hypothesis part
            xmlNodePtr hypothesisNode = 0;
            if (!hypothesisNode)
            {
                hypothesisNode = getNode("HistogramWeakHypothesis", node);
                if (hypothesisNode)
                {
                    tmpPredict.push_back(vector<float>());
                    loadHistogramWeakHypothesis(hypothesisNode, tmpPredict.back());
                }
                else 
                {
                    hypothesisNode = getNode("DecisionTreeWeakHypothesis", node);
                    if (hypothesisNode)
                    {
                        tmpPredict.push_back(vector<float>());
                        loadDTWeakHypothesis(hypothesisNode, tmpPredict.back());
                    }
                }
            }

            if (!hypothesisNode)
            {
                cerr << "Warning: Incomplete stage encountered" << endl;
            }
            else
            {
                loadHaarFeature(hypothesisNode, &stage);
            }
            //cout << stage.x << ";" << stage.y << "]" << endl;
            tmpStages.push_back(stage);            
        }
    }

    // Alloc the stages and alphas, and copy loaded data
    classifier->stageCount = tmpStages.size();
    classifier->evaluatedStageCount = tmpStages.size();
    classifier->alphaCount = tmpPredict[0].size();
   
    classifier->stage = new THaarStage[tmpStages.size()];
    classifier->alpha = new float[tmpStages.size() * classifier->alphaCount];

    for (unsigned s = 0; s < tmpStages.size(); ++s)
    {
        ((THaarStage*)(classifier->stage))[s] = tmpStages[s];
        //THaarStage & stage = ((THaarStage*)(classifier->stage))[s];
        //printf("%d: tp=%d; x=%d; y=%d; w=%d; h=%d; theta=%f; interval=<%f;%f>; %d bins\n",
        //        s, int(stage.tp), stage.x, stage.y, stage.w, stage.h, stage.theta_b, stage.min, stage.max, stage.bins);
        if (tmpPredict[s].size() != classifier->alphaCount)
            cerr << "Warning: Incorrect stage (" << s << ")" << endl;
        float * dstAlpha = classifier->alpha + classifier->alphaCount * s;
        copy(tmpPredict[s].begin(), tmpPredict[s].end(), dstAlpha);
    }
}


/// Strings with types of classifiers (Indexed by ClassifierType).
const char *const classifierTypeStrings[] = {
    "UNKNOWN",
    "LRD",
    "LRP",
    "LRP3"
    "HAAR"
    "LBP",
};

// EXPORT CRAP
// FIXME modify to export all types of classifiers

void exportClassifierHeader(TClassifier * c, ostream & str, const char * name)
{
    assert(sizeof(classifierTypeStrings) == (size_t)numClassifierTypes*sizeof(*classifierTypeStrings));

    str << "#ifndef _"<< name << "_\n#define _" << name << "_\n\n";

    str << "#include <classifier.h>\n\n";

    str << "#define STAGE_COUNT " << c->stageCount << "\n";
    str << "#define ALPHA_COUNT " << c->alphaCount << "\n\n";

    switch (c->tp)
    {
    case LRD:
    case LRP:
    case LRP3:
    case LBP:
        {
            str << "TStage _stages_" << name << "[STAGE_COUNT] = {\n";
            for (unsigned s = 0; s < c->stageCount; ++s)
            {
                TStage & stg = ((TStage*)(c->stage))[s];
                str << "{ " <<
                    stg.x << ", " << stg.y << ", " <<
                    stg.w << ", " << stg.h << ", " <<
                    int(stg.A) << ", " << int(stg.B) << ", " << int(stg.C) << ", " <<
                    showpoint << stg.theta_b << "f, " <<
                    0 << ", " << // alpha
                    0 << ", " << 0 << "},\n"; // szType, posType
            }
            str << "};\n\n";
        } // case LRD
        break;
    case HAAR:
        {
            str << "THaarStage _stages_" << name << "[STAGE_COUNT] = {\n";
            for (unsigned s = 0; s < c->stageCount; ++s)
            {
                THaarStage & stg = ((THaarStage*)(c->stage))[s];
                str << "{ " <<
                    stg.x << ", " << stg.y << ", " <<
                    stg.w << ", " << stg.h << ", " << stg.tp << ", " <<
                    showpoint << stg.min << "f, " << stg.max << "f, " << stg.bins << ", " <<
                    stg.theta_b << "f, " <<
                    0 << ", " << // alpha
                    "},\n";
            }
            str << "};\n\n";
        } // case HAAR
        break;
    default:
        str << "// Unknown classifier type.\n\n";
        break;
    }

    str << "float _alphas_" << name << "[STAGE_COUNT * ALPHA_COUNT] = {\n";
    float * alpha = c->alpha;
    for (unsigned s = 0; s < c->stageCount; ++s, alpha+=c->alphaCount)
    {
        for (unsigned a = 0; a < c->alphaCount; ++a)
        {
            str << alpha[a] << "f, ";
        }
        str << "\n";
    }
    str << "};\n\n";
    
    str << "TClassifier " << name << " = {\n";
    //str << "(ClassifierType)" << int(c->tp) << ", STAGE_COUNT, STAGE_COUNT, ALPHA_COUNT, " << 
    str << classifierTypeStrings[c->tp] << ", STAGE_COUNT, STAGE_COUNT, ALPHA_COUNT, " << 
        c->threshold << ", " << 
        c->width << ", " << c->height << ", " <<
        "(TStage*) _stages_" << name << ", (float*) _alphas_" << name << ", \n";
    str << "};\n\n";
    
    str << "#endif\n" << endl;
}


void exportClassifierTemplate(TClassifier * c, ostream & str, const char * name)
{

    for (unsigned s = 0; s < c->stageCount; s+=4)
    {
        for (int i = 0; i < 4; ++i)
        {
            TStage & stg = ((TStage*)(c->stage))[s+i];
            int prev = i - 1;
            if (i == 0) prev = 3;
            str << "R["<< i <<"] = R["<< prev <<"] + evalLRDStageSimple<"
                << stg.w << "," << stg.h
                << ">(image,smpOffset,((TStage*)c.stage) + " << i  << ");" << endl;
        }        
        
        for (int i = 0; i < 4; ++i)
        {
            TStage & stg = ((TStage*)(c->stage))[s+i];
            if (stg.theta_b > -50)
            {
                str << "if (R["<< i <<"] < " << stg.theta_b << ") return 0;\n";
            }
        }
        str << endl;
    }
}


void exportStaticClassifierCode(TClassifier * c, ostream & str, const char * name)
{
    str << "#include <feature_macros.h>\n\n";

    str << "#define STAGE_COUNT " << c->stageCount << "\n";
    str << "#define ALPHA_COUNT " << c->alphaCount << "\n\n";
    
    // Write table with alphas
    str << "float _alphas_" << name << "[STAGE_COUNT * ALPHA_COUNT] = {\n";
    float * alpha = c->alpha;
    for (unsigned s = 0; s < c->stageCount; ++s, alpha+=c->alphaCount)
    {
        for (unsigned a = 0; a < c->alphaCount; ++a)
        {
            str << alpha[a] << "f, ";
        }
        str << "\n";
    }
    str << "};\n\n";

    // Write evaluation macro
    str << "#define EVAL_" << name << "_CLASSIFIER(data,ls,response)\\\n";
    
    for (unsigned s = 0; s < c->stageCount; ++s)
    {
        switch (c->tp)
        {
        case LRP:
            {
                TStage & stg = ((TStage*)(c->stage))[s];
                str << "LRP(data,ls," << stg.x << "," << stg.y << "," << stg.A%3 << "," << stg.A/3 << ","
                    << stg.B%3 << "," << stg.B/3 << "," << stg.w << "," << stg.h << ",px" << stg.w << stg.h << ","
                    << "((float*)_alphas_" << name << "+" << c->alphaCount * s << "), response)\\\n";
            }
            break;
        case LBP:
            {
                TStage & stg = ((TStage*)(c->stage))[s];
                str << "LBP(data,ls," << stg.x << "," << stg.y << ","
                    << stg.w << "," << stg.h << ",px" << stg.w << stg.h << ","
                    << "((float*)_alphas_" << name << "+" << c->alphaCount * s << "), response)\\\n";
            }
            break;
        default:
            break;
        }
    }
}

