/*
 * xml2h.cpp
 * $Id: xml2h.cpp 99 2009-04-29 10:04:45Z ihavel $
 * XML to .h converter - Converts XML classifiers from
 * Research Boosting Framework to .h file so they can
 * be statically linked to programs.
 *
 * xml2h classifier.xml -o test_cls.h -n test
 * -o filename      Output file
 * -n identifier    Classifier identifier in the code
 */

#include <classifier.h>

#include <stdio.h>
#include <iostream>
#include <fstream>


using namespace std;

void exportClassifierTemplate(TClassifier * c, ostream & str, const char * name);

int main(int argc, char ** argv)
{
    const char * output = 0; // -o
    const char * name = 0;   // -n
    const char * input = 0;

    for (int arg = 1; arg < argc; ++arg)
    {
        if (strcmp(argv[arg], "-o") == 0)
        {
            output = argv[++arg];
            continue;
        }
        if (strcmp(argv[arg], "-n") == 0)
        {
            name = argv[++arg];
            continue;
        }
        input = argv[arg];
    }

    if (!name || !input || !output)
    {
        cerr << "xml2h -o output.h -n name input.xml" << endl;
        return 0;
    }
    
    ofstream fout(output);

    TClassifier * classifier = loadClassifierXML(input);

    if (classifier)
    {
        initClassifier(classifier);
        // use classifier
        //exportClassifierHeader(classifier, fout, name);
        exportStaticClassifierCode(classifier, fout, name);
        //exportClassifierTemplate(classifier, fout, name);
        releaseClassifier(&classifier);
    }
    else
    {
        cerr << "Cannot load '" << input << "'" << endl;
    }

    return 0;
}
